# -*- coding: utf-8 -*-

from __future__ import print_function
import sys
from six import text_type
from jinja2 import Environment, nodes
from jinja2.ext import Extension
from jinja2.exceptions import TemplateSyntaxError

from .text import RunX, Segment

class Env(Environment):

    def handle_exception(self, *args, **kwargs):
        exc_type, exc_value, tb = sys.exc_info()
        red_fmt = '\033[31m%s\033[0m'
        blue_fmt = '\033[34m%s\033[0m'
        error_type = red_fmt % ('error type:  %s' % exc_type)
        error_message = red_fmt % ('error message:  %s' % exc_value)
        print(error_type)
        print(error_message)
        if exc_type is TemplateSyntaxError:
            lineno = exc_value.lineno
            source = kwargs['source']
            src_lines = source.splitlines()
            for i, line in enumerate(src_lines):
                if i + 1 == lineno:
                    line_str = red_fmt % ('line %d : %s' % (i + 1, line))
                elif i + 1 in [lineno - 1, lineno + 1]:
                    line_str = blue_fmt % ('line %d : %s' % (i + 1, line))
                else:
                    line_str = 'line %d : %s' % (i + 1, line)
                print(line_str)
        Environment.handle_exception(self, *args, **kwargs)

class NodeExtension(Extension):
    tags = set(['node', 'root', 'body', 'para', 'run', 'hyperlink',
                'table', 'row', 'cell', 'default', 'headtail'])

    def __init__(self, environment):
        super(self.__class__, self).__init__(environment)
        environment.extend(root = None)

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = []
        return nodes.CallBlock(self.call_method('_node', args),
                               [], [], body).set_lineno(lineno)

    def _node(self, key, caller):
        node = self.environment.root.get_node(key)
        return key

class SegmentExtension(Extension):
    tags = set(['seg'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = parser.parse_statements(['name:endseg'], drop_needle=True)
        return nodes.CallBlock(self.call_method('_seg', args),
                               [], [], body).set_lineno(lineno)

    def _seg(self, key, caller):
        node = self.environment.root.get_node(key)
        rv = caller()
        rv = node.process_rv(rv)
        return rv

class PicExtension(Extension):
    tags = set(['pic', 'img'])

    def __init__(self, environment):
        super(self.__class__, self).__init__(environment)
        environment.extend(root = None)

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = []
        return nodes.CallBlock(self.call_method('_node', args),
                               [], [], body).set_lineno(lineno)

    def _node(self, fname, caller):
        node = self.environment.root.current_node
        fname = text_type(fname)
        if isinstance(node, RunX):
            node._parent.replace_pic(fname)
        elif isinstance(node, Segment):
            node._parent._parent.replace_pic(fname)
        return fname


# -*- coding: utf-8 -*-

from __future__ import print_function
from six import text_type
from unicodedata import normalize
from copy import deepcopy
from docx.document import _Body
from .utils import TreeProperty

node_clses = {}
class Node(object):
    node_map = TreeProperty('node_map')
    ext_tag = 'node'

    def __init__(self):
        self._children = []
        self.unpacked = False
        self.no = 0

    @property
    def depth(self):
        if not hasattr(self, '_depth'):
            if not hasattr(self, '_parent') or self._parent is self:
                self._depth = 0
            else:
                self._depth = self._parent.depth + 1
        return self._depth

    @property
    def node_key(self):
        return '%s,%s' % (self._parent.node_key, self.no)

    @property
    def node_tag(self):
        return "{%%%s '%s' %%}" % (self.ext_tag, self.node_key)

    @property
    def print_tag(self):
        return self.node_tag

    def children_to_tag(self):
        x = []
        for child in self._children:
            x.append(child.to_tag())
        return '\n'.join(x)

    def to_tag(self):
        self.node_map[self.node_key] = self
        if self._children:
            return self.children_to_tag()
        return self.node_tag

    def tag_tree(self):
        print_tag = normalize("NFKD", text_type(self.print_tag))
        print('\t' * self.depth, print_tag)
        for child in self._children:
            child.tag_tree()

    def get_node_cls(self, sub_element):
        return node_clses.get(type(sub_element), Default)

    def unpack_element(self):
        for sub_element in self._element:
            node_cls = self.get_node_cls(sub_element)
            child = node_cls(sub_element, self)
            self.add_child(child)
            if not self.unpacked and child.unpacked:
                self.unpacked = True

    def clear_element(self):
        if not self.unpacked:
            del self._children[:]
            return
        elms = self._element[:]
        for elm in elms:
            self._element.remove(elm)

    def unpack_and_clear(self):
        self.unpack_element()
        self.clear_element()

    def copy_element(self):
        return deepcopy(self._element)

    def add_child(self, child):
        child.no = len(self._children)
        #child._parent = self
        self._children.append(child)

    def enter(self):
        pass

    def reenter(self):
        self.enter()

    def child_reenter(self):
        pass

    def exit(self):
        pass

    def process_child_rv(self, rv):
        pass

    def __str__(self):
        return self.__class__.__name__ +' , ' + self.node_tag

class HtNode(Node):
    ext_tag = 'headtail'

    def __init__(self, parent):
        self._parent = parent
        Node.__init__(self)

    def to_tag(self):
        if self.no == 0:
            return ''
        else:
            return Node.to_tag(self)

    def enter(self):
        if self.no != 0:
            self._parent.exit()

    def exit(self):
        if self.no == 0:
            self._parent.enter()


class RvNode(Node):
    ext_tag = 'rv'

    def enter(self):
        self.rv = self.copy_element()

    def exit(self):
        self._parent.process_child_rv(self.rv)

    def process_child_rv(self, rv):
        self.rv.append(rv)

class Root(RvNode):
    ext_tag = 'root'

    def __init__(self, root):
        self._root = root
        self._element = root._element
        self._parent = self
        self.part = root.part
        self.node_map = {}
        Node.__init__(self)
        self.head_node = head_node = HtNode(self)
        tail_node = HtNode(self)
        self.add_child(head_node)
        self.unpack_element()
        self.add_child(tail_node)
        self.current_node = head_node
        self.current_key = ''

    @property
    def node_key(self):
        return '0'

    def enter(self):
        self.clear_element()
        self.rv = self._element

    def exit(self):
        self.current_node = self.head_node
        self.current_key = ''

    def find_lca(self, pre, next):
        # find lowest common ancestor
        next_branch = []
        if pre.depth > next.depth:
            for i in range(next.depth, pre.depth):
                pre.exit()
                # print(pre, 'pre up', pre._parent)
                pre = pre._parent

        elif pre.depth < next.depth:
            for i in range(pre.depth, next.depth):
                next_branch.insert(0, next)
                # print(next, 'next up', next._parent)
                next = next._parent
        if pre is next:
            pass
        else:
            pre_parent = pre._parent
            next_parent = next._parent
            while pre_parent != next_parent:
                # print(pre, next, 'up together')
                pre.exit()
                pre = pre_parent
                pre_parent = pre._parent
                next_branch.insert(0, next)
                next = next_parent
                next_parent = next._parent
            pre.exit()
            if pre_parent._children.index(pre) > pre_parent._children.index(next):
                pre_parent.child_reenter()
                next.reenter()
            else:
                next.enter()

        for next in next_branch:
            next.enter()

    def get_node(self, key):
        #if self.current_key == key:
        #    return self.current_node
        self.last_node = self.current_node
        self.last_key = self.current_key
        self.current_node = self.node_map.get(key)
        self.current_key = key
        self.find_lca(self.last_node, self.current_node)
        return self.current_node


class BodyX(_Body, RvNode):
    ext_tag= 'body'

    def __init__(self, element, parent):
        _Body.__init__(self, element, parent)
        RvNode.__init__(self)
        self.unpack_and_clear()


class Default(RvNode):
    ext_tag = 'default'

    def __init__(self, element, parent):
        RvNode.__init__(self)
        self._element = element
        self._parent = parent


# -*- coding: utf-8 -*-

from docx.oxml import CT_Body, CT_Tbl, CT_Row, CT_Tc, CT_R
from docx.oxml import register_element_cls

from .node import BodyX, node_clses
from .text import ParagraghX, RunX, HyperlinkX
from .text import CT_Para, CT_Run, CT_Hyperlink, CT_Drawing
from .table import TableX, RowX, CellX

register_element_cls('w:p', CT_Para)
register_element_cls('w:hyperlink', CT_Hyperlink)
register_element_cls('w:r', CT_Run)
register_element_cls('w:drawing', CT_Drawing)

def register_node_cls(xml_cls, cls):
    node_clses[xml_cls] = cls

register_node_cls(CT_Body, BodyX)
register_node_cls(CT_Para, ParagraghX)
register_node_cls(CT_Run, RunX)
register_node_cls(CT_Tbl, TableX)
register_node_cls(CT_Row, RowX)
register_node_cls(CT_Tc, CellX)
register_node_cls(CT_Hyperlink, HyperlinkX)

from .writer import DocxWriter



# -*- coding: utf-8 -*-

from __future__ import print_function
from docx.oxml import CT_P, CT_R
from docx.oxml.xmlchemy import BaseOxmlElement, ZeroOrOne, ZeroOrMore, OneAndOnlyOne

class CT_Drawing(BaseOxmlElement):
    inline = OneAndOnlyOne('wp:inline')

class CT_Run(CT_R):
    drawing = ZeroOrOne('w:drawing')

class CT_Hyperlink(BaseOxmlElement):
    r = ZeroOrMore('w:r')

class CT_Para(CT_P):
    hyperlink = ZeroOrMore('w:hyperlink')

from docx.text.paragraph import Paragraph
from docx.text.run import Run
from docx.shared import Parented

from .node import RvNode
from .utils import control_split, var_split, tag_test, fix_test, tag_fix

class ParagraghX(Paragraph, RvNode):
    ext_tag = 'para'

    def __init__(self, element, parent):
        Paragraph.__init__(self, element, parent)
        RvNode.__init__(self)
        self.unpack_p()
        self.current_drawing = None

    def unpack_p(self):
        if not tag_test(self.text + self.hyperlink_text):
            return
        self.fix_and_unpack()

    def fix_and_unpack(self):
        text_4_fix = self.text_4_fix()
        if fix_test(text_4_fix):
            fixed = tag_fix(text_4_fix)
            #print(text_4_fix)
            #print(fixed)
            for i, sub_element in enumerate(self._element):
                if i in fixed:
                    text = fixed[i]
                    if text == '':
                        if sub_element.text != '':
                            continue
                        #else:#drawing
                    else:
                        sub_element.text = text
                node_cls = self.get_node_cls(sub_element)
                child = node_cls(sub_element, self)
                self.add_child(child)
            self.unpacked = True
            self.clear_element()
        else:
            RvNode.unpack_and_clear(self)


    def register_drawing(self, drawing):
        self.current_drawing = drawing

    def replace_pic(self, fname):
        if self.current_drawing is None:
            return
        try:
            rId, image = self.part.get_or_add_image(fname)
        except:
            return
        inline = self.current_drawing.inline
        inline.docPr.id = self.part.next_id
        inline.docPr.name = fname
        inline.docPr.descr = fname
        pic = inline.graphic.graphicData.pic
        pic.nvPicPr.cNvPr.id = 0#pic_id
        pic.nvPicPr.cNvPr.name = fname
        pic.blipFill.blip.embed = rId

    @property
    def print_tag(self):
        return self.node_tag + '  ' + self.text

    @property
    def hyperlinks(self):
        return [Hyperlink(hyperlink, self) for hyperlink in self._p.hyperlink_lst]

    @property
    def hyperlink_text(self):
        text = ''
        for hyperlink in self.hyperlinks:
            text += hyperlink.text
        return text

    def text_4_fix(self):
        text = ''
        tmpl = '___%d___'
        for i,elm in enumerate(self._element):
            if isinstance(elm, CT_R):
                text += (tmpl % i ) + elm.text
        return text


class RunX(Run, RvNode):
    ext_tag = 'run'

    def __init__(self, element, parent):
        Run.__init__(self, element, parent)
        RvNode.__init__(self)
        self.unpack_r()

    def unpack_r(self):
        if not tag_test(self.text):
            return
        self.unpacked = True
        parts = control_split(self.text)
        for index,part in enumerate(parts):
            if part == '':
                continue
            if index % 2 == 0:
                sub_parts = var_split(part)
                for sub_index, sub_part in enumerate(sub_parts):
                    if sub_part == '':
                        continue
                    if sub_index % 2 == 0:
                        child = TextSegment(sub_part, self)
                    else:
                        child = VarSegment(sub_part, self)
                    self.add_child(child)
            else:
                child = ControlSegment(part, self)
                self.add_child(child)

    def enter(self):
        self.rv = self.copy_element()
        if self.rv.drawing is not None:
            self._parent.register_drawing(self.rv.drawing)
        self.children_rvs = ''

    def process_child_rv(self, rv):
        self.children_rvs += rv

    def exit(self):
        if self._children:
            self.rv.text = self.children_rvs
        self._parent.process_child_rv(self.rv)

    @property
    def print_tag(self):
        return self.node_tag + '  ' + self.text

class Segment(RvNode):

    def __init__(self, text, parent):
        RvNode.__init__(self)
        self.text = text
        self._parent = parent

    def process_rv(self, rv):
        self.rv = rv
        return rv

    def enter(self):
        self.rv = ''

class TextSegment(Segment):

    @property
    def node_tag(self):
        tmpl = "{%%seg '%s'%%}{%%endseg%%}"
        return tmpl % self.node_key

    def process_rv(self, rv):
        self.rv = self.text
        return self.text

class VarSegment(Segment):

    @property
    def node_tag(self):
        tmpl = "{%%seg '%s'%%}%s{%%endseg%%}"
        return tmpl % (self.node_key, self.text)

class ControlSegment(Segment):

    @property
    def node_tag(self):
        tmpl = "%s{%%seg '%s'%%}{%%endseg%%}"
        return tmpl % (self.text, self.node_key)


class Hyperlink(Parented):

    def __init__(self, hyperlink, parent):
        super(Hyperlink, self).__init__(parent)
        self._hyperlink = self._element = hyperlink

    def add_run(self, text=None, style=None):
        r = self._hyperlink.add_r()
        run = Run(r, self)
        if text:
            run.text = text
        if style:
            run.style = style
        return run

    def clear(self):
        self._hyperlink.clear_content()
        return self

    @property
    def runs(self):
        return [Run(r, self) for r in self._hyperlink.r_lst]

    @property
    def text(self):
        text = ''
        for run in self.runs:
            text += run.text
        return text

    @text.setter
    def text(self, text):
        self.clear()
        self.add_run(text)

from .utils import bind
class HyperlinkX(Hyperlink, RvNode):
    ext_tag = 'hyperlink'

    def __init__(self, hyperlink, parent):
        Hyperlink.__init__(self, hyperlink, parent)
        RvNode.__init__(self)
        bind(ParagraghX.fix_and_unpack, self)
        bind(ParagraghX.text_4_fix, self)
        bind(ParagraghX.register_drawing, self)
        bind(ParagraghX.replace_pic, self)
        self.unpack_hyperlink()
        self.current_drawing = None

    def unpack_hyperlink(self):
        if not tag_test(self.text):
            return
        self.fix_and_unpack()

    @property
    def print_tag(self):
        return self.node_tag + '   ' + self.text



# -*- coding: utf-8 -*-

from __future__ import print_function
from docx import Document
from docx.opc.constants import RELATIONSHIP_TYPE as RT
from .node import Root
from .ext import Env, NodeExtension, SegmentExtension, PicExtension

class DocxWriter(object):

    reltypes = [RT.HEADER, RT.FOOTER]

    def __init__(self, fname, debug=False):
        self.debug = debug
        self.load(fname)

    def load(self, fname):
        self.prepare_env()
        self.tpls = []
        self.roots = []
        self.document = Document(fname)
        for rel in self.document.part.rels.values():
            if rel.reltype in self.reltypes:
                root = Root(rel._target)
                if root.unpacked:
                    self.roots.append(root)
        root = Root(self.document)
        if root.unpacked:
            self.roots.append(root)
        for root in self.roots:
            tpl_source = root.to_tag()
            if self.debug:
                root.tag_tree()
                print(tpl_source)
            jinja_tpl = self.jinja_env.from_string(tpl_source)
            self.tpls.append((root, jinja_tpl, tpl_source))

    def prepare_env(self):
        self.jinja_env = Env(extensions=[NodeExtension, SegmentExtension, PicExtension])

    def render(self, payload):
        for root,jinja_tpl,tpl_source in self.tpls:
            self.jinja_env.root = root
            rv = jinja_tpl.render(payload)

    def save(self, fname):
        self.document.save(fname)


# -*- coding: utf-8 -*-
import re

BLOCK_START_STRING = '{%'
BLOCK_END_STRING = '%}'
VARIABLE_START_STRING = '{{'
VARIABLE_END_STRING = '}}'
TAGTEST = '%s.+%s|%s.+%s' % (BLOCK_START_STRING, BLOCK_END_STRING, VARIABLE_START_STRING, VARIABLE_END_STRING)
XVTEST = '^ *%s *xv.+%s *$' % (BLOCK_START_STRING, BLOCK_END_STRING)
BLOCKTEST = '%s.+%s' % (BLOCK_START_STRING, BLOCK_END_STRING)
BLOCKMIX = '({%(?:(?!%}).)+?)(___\d*___)(.+?%})'# % (BLOCK_START_STRING, BLOCK_END_STRING)
CONTROLSPLIT = '((?:%s.+?%s)+)' % (BLOCK_START_STRING, BLOCK_END_STRING)
CONTROLSPLIT2 = '((?:%s(?:(?!yn ).)+?%s)+)' % (BLOCK_START_STRING, BLOCK_END_STRING)
VARSPLIT = '((?:%s.+?%s)+)' % (VARIABLE_START_STRING, VARIABLE_END_STRING)

FIXTEST = '({(?:___\d+___)?(?:{|%).+?(?:}|%)(?:___\d+___)?})'
RUNSPLIT = '(___\d+___)'
RUNSPLIT2 = '___(\d+)___'

def tag_test(txt):
    p = re.compile(TAGTEST)
    rv = p.findall(txt)
    return bool(rv)

def block_tag_test(txt):
    p = re.compile(BLOCKTEST)
    rv = p.findall(txt)
    return bool(rv)

def control_split(txt):
    split_pattern = re.compile(CONTROLSPLIT)
    parts = split_pattern.split(txt)
    return parts

def var_split(txt):
    split_pattern = re.compile(VARSPLIT)
    parts = split_pattern.split(txt)
    return parts

#need a better way to handle this
def fix_test(txt):
    split_pattern = re.compile(FIXTEST)
    parts = split_pattern.split(txt)
    for i, part in enumerate(parts):
        if i % 2 == 1:
            pattern = re.compile(RUNSPLIT)
            rv = pattern.findall(part)
            if rv :
                return True

def tag_fix(txt):
    split_pattern = re.compile(FIXTEST)
    parts = split_pattern.split(txt)
    p = ''
    for i, part in enumerate(parts):
        if i % 2 == 1:
            p += fix_step2(part)
        else:
            p += part
    split_pattern2 = re.compile(RUNSPLIT2)
    parts = split_pattern2.split(p)
    d = {}
    for i in range(1, len(parts), 2):
            d[int(parts[i])] = parts[i + 1]
    return d

def fix_step2(txt):
    split_pattern = re.compile(RUNSPLIT)
    parts = split_pattern.split(txt)
    p0 = ''
    p1 = ''
    for index,part in enumerate(parts):
        if index % 2 == 0:
            p0 += part
        else:
            p1 += part
    return p0+p1

class TreeProperty(object):

    def __init__(self, name):
        self.name = name
        self._name = '_' + name

    def __set__(self, instance, value):
        instance.__dict__[self._name] = value

    def __get__(self, instance, cls):
        if not hasattr(instance, self._name):
            instance.__dict__[self._name] = getattr(instance._parent, self.name)
        return instance.__dict__[self._name]

from types import MethodType
def bind(method, instance):
    if isinstance(method, MethodType):
        instance.__dict__[method.__name__] = MethodType(method.im_func, instance)
    else:
        instance.__dict__[method.__name__] = MethodType(method, instance)


# -*- coding: utf-8 -*-

from docx.table import Table, _Row, _Cell
from .node import RvNode

class TableX(Table, RvNode):
    ext_tag = 'table'

    def __init__(self, element, parent):
        Table.__init__(self, element, parent)
        RvNode.__init__(self)
        self.unpack_table()

    def unpack_table(self):
        self.unpack_and_clear()

class RowX(_Row, RvNode):
    ext_tag = 'row'

    def __init__(self, element, parent):
        _Row.__init__(self, element, parent)
        RvNode.__init__(self)
        self.unpack_row()

    def unpack_row(self):
        self.unpack_and_clear()

    def child_reenter(self):
        self.exit()
        self.enter()

class CellX(_Cell, RvNode):
    ext_tag = 'cell'

    def __init__(self, element, parent):
        _Cell.__init__(self, element, parent)
        RvNode.__init__(self)
        self.unpack_cell()

    def unpack_cell(self):
        self.unpack_and_clear()

class Range(object):
    pass




import os
from io import open
from setuptools import setup

CUR_DIR = os.path.abspath(os.path.dirname(__file__))
README = os.path.join(CUR_DIR, "README_en.md")
with open(README, 'r', encoding='utf-8') as fd:
    long_description = fd.read()

setup(
    name = 'pydocxtpl',
    version = "0.2.1",
    author = 'Zhang Yu',
    author_email = 'zhangyu836@gmail.com',
    url = 'https://github.com/zhangyu836/python-docx-templater',
    packages = ['pydocxtpl'],
	install_requires = ['python-docx >= 0.8.10', 'jinja2', 'six'],
    description = ( 'A python module to generate docx files from a docx template' ),
    long_description = long_description,
    long_description_content_type = "text/markdown",
    platforms = ["Any platform "],
    license = 'MIT',
    keywords = ['Word', 'docx', 'template']
)